@@ -53,9 +53,15 @@
 #include "havealloca.h"
 #include "havesetresuid.h"
 
+#if !defined(__OPTIMIZE__) && defined(__linux__)
+#include <sys/prctl.h>
+#endif
+
 char serverroot[1024];
 
+#ifdef SUPPORT_MULTIPROC
 unsigned long instances=1;
+#endif
 unsigned long timeout_secs=23;
 tai6464 now,next;
 
@@ -85,49 +91,6 @@ int64 new_io_timeouted() {
 #define io_timeouted new_io_timeouted
 #endif
 
-#ifdef DEBUG_EVENTS
-void new_io_wantwrite(int64 s,const char* file,unsigned int line) {
-  char a[FMT_ULONG];
-  char b[FMT_ULONG];
-  a[fmt_ulong(a,s)]=0;
-  b[fmt_ulong(b,line)]=0;
-  buffer_putmflush(buffer_2,"DEBUG: ",file,":",b,": io_wantwrite(",a,")\n");
-  io_wantwrite(s);
-}
-
-void new_io_dontwantwrite(int64 s,const char* file,unsigned int line) {
-  char a[FMT_ULONG];
-  char b[FMT_ULONG];
-  a[fmt_ulong(a,s)]=0;
-  b[fmt_ulong(b,line)]=0;
-  buffer_putmflush(buffer_2,"DEBUG: ",file,":",b,": io_dontwantwrite(",a,")\n");
-  io_dontwantwrite(s);
-}
-
-void new_io_wantread(int64 s,const char* file,unsigned int line) {
-  char a[FMT_ULONG];
-  char b[FMT_ULONG];
-  a[fmt_ulong(a,s)]=0;
-  b[fmt_ulong(b,line)]=0;
-  buffer_putmflush(buffer_2,"DEBUG: ",file,":",b,": io_wantread(",a,")\n");
-  io_wantread(s);
-}
-
-void new_io_dontwantread(int64 s,const char* file,unsigned int line) {
-  char a[FMT_ULONG];
-  char b[FMT_ULONG];
-  a[fmt_ulong(a,s)]=0;
-  b[fmt_ulong(b,line)]=0;
-  buffer_putmflush(buffer_2,"DEBUG: ",file,":",b,": io_dontwantread(",a,")\n");
-  io_dontwantread(s);
-}
-
-#define io_wantwrite(s) new_io_wantwrite(s,__FILE__,__LINE__)
-#define io_wantread(s) new_io_wantread(s,__FILE__,__LINE__)
-#define io_dontwantwrite(s) new_io_dontwantwrite(s,__FILE__,__LINE__)
-#define io_dontwantread(s) new_io_dontwantread(s,__FILE__,__LINE__)
-#endif
-
 const char months[] = "JanFebMarAprMayJunJulAugSepOctNovDec";
 
 #ifdef SUPPORT_CGI
@@ -273,7 +236,7 @@ void cleanup(int64 fd) {
 #endif
 #ifdef SUPPORT_HTTPS
     if (h->t==HTTPSREQUEST || h->t==HTTPSPOST || h->t==HTTPSACCEPT ||
-	h->t==HTTPSRESPONSE) --https_connections;
+	h->t==HTTPSACCEPT_CHECK || h->t==HTTPSRESPONSE) --https_connections;
 #endif
 
 #if defined(SUPPORT_FTP)
@@ -293,8 +256,13 @@ void cleanup(int64 fd) {
     free(h->ftppath);
 #endif
 #ifdef SUPPORT_HTTPS
+#ifdef USE_OPENSSL
     if (h->ssl) SSL_free(h->ssl);
 #endif
+#ifdef USE_POLARSSL
+    ssl_free(&h->ssl);
+#endif
+#endif
 #ifdef SUPPORT_SMB
     close_all_handles(&h->h);
 #endif
@@ -308,7 +276,7 @@ void cleanup(int64 fd) {
   }
 }
 
-size_t header_complete(struct http_data* r) {
+size_t header_complete(struct http_data* r,int64 sock) {
   long i;
   long l=array_bytes(&r->r);
   const char* c=array_start(&r->r);
@@ -321,6 +289,34 @@ size_t header_complete(struct http_data* r) {
 #endif
      )
   {
+    /* first of all, in case someone is bombarding me with crap, detect
+     * that early so we can drop the connection */
+    if (l>10) {
+      tai6464 tarpit;
+      for (i=0; i<l; ++i) {
+	if (c[i]==' ') goto ok;
+	if (c[i]<'A' || c[i]>'Z') break;
+      }
+      /* detected invalid HTTP request */
+      if (logging) {
+	char buf[100];
+	buffer_puts(buffer_1,"close/not_http_traffic ");
+	buffer_putulong(buffer_1,sock);
+	buffer_putspace(buffer_1);
+	buffer_put(buffer_1,buf,fmt_ip6c(buf,r->peerip));
+	buffer_putnlflush(buffer_1);
+      }
+      io_dontwantread(sock);
+      io_dontwantwrite(sock);
+      --http_connections;
+      changestate(r,PUNISHMENT);
+      tarpit=now;
+      tarpit.sec.x+=10;
+      io_timeout(sock,tarpit);
+      return 0;
+    }
+ok:
+
     /* Erdgeist nudged me into optimizing this :-)
      * I'd be surprised if this optimization has any measurable
      * advantage, but it sure is impressive */
@@ -344,7 +340,10 @@ size_t header_complete(struct http_data* r) {
      * byte 0: 0 ("session message"),
      * bytes 1-3: message length (big endian) */
     uint32 len;
-    if (c[0]!=0) return 1;
+    if (c[0]!=0) {
+//      printf("can't happen error: netbios byte 0 was not 0!?\n");
+      return 1;
+    }
     len=uint32_read_big(c) & 0x00ffffff;
     if (l>=len+4) return len+4;
 #endif
@@ -445,9 +444,9 @@ static uid_t __uid;
 static gid_t __gid;
 
 static int prepare_switch_uid(const char* new_uid) {
+  uid_t u=0;
+  gid_t g=0;
   if (new_uid) {
-    uid_t u=0;
-    gid_t g=0;
     if (new_uid[0]>='0' && new_uid[0]<='9') {
       unsigned long l;
       const char *c=new_uid+scan_ulong(new_uid,&l);
@@ -467,20 +466,30 @@ static int prepare_switch_uid(const char* new_uid) {
     }
     __uid=u;
     __gid=g;
+  } else {
+    /* default to chown(nobody) and chgrp(nogroup) */
+    __uid=-1;
+    __gid=-1;
   }
   return 0;
 }
 
 int switch_uid() {
 #ifdef LIBC_HAS_SETRESUID
-  if (setresgid(__gid,__gid,__gid)) return -1;
-  if (setgroups(1,&__gid)) return -1;
-  if (setresuid(__uid,__uid,__uid)) return -1;
+  if (setresgid(__gid,__gid,__gid) ||
+      setgroups(1,&__gid) ||
+      setresuid(__uid,__uid,__uid))
 #else
-  if (setgid(__gid)) return -1;
-  if (setgroups(1,&__gid)) return -1;
-  if (setuid(__uid)) return -1;
+  if (setgid(__gid) ||
+      setgroups(1,&__gid) ||
+      setuid(__uid))
 #endif
+  {
+    /* if dropping privileges failed with EPERM, we are already running
+     * with reduced privileges.  Don't make a fuss about it then. */
+    if (errno==EPERM) return 0;
+    return -1;
+  }
   return 0;
 }
 #endif
@@ -523,6 +532,8 @@ char* sshd;
 unsigned long ssh_timeout;
 #endif
 
+int limit_to_lan;
+
 static void accept_server_connection(int64 i,struct http_data* H,unsigned long ftptimeout_secs,tai6464 nextftp) {
   /* This is an FTP or HTTP(S) or SMB server connection.
     * This read event means that someone connected to us.
@@ -566,8 +577,58 @@ static void accept_server_connection(int64 i,struct http_data* H,unsigned long f
       buffer_putulong(buffer_1,port);
       buffer_puts(buffer_1," ");
       buffer_putulong(buffer_1,connections-1);
+      buffer_puts(buffer_1," ");
+      {
+	const char* service;
+	switch (H->t) {
+	case HTTPSERVER4:
+	case HTTPSERVER6: service="http"; break;
+#ifdef SUPPORT_HTTPS
+	case HTTPSSERVER4:
+	case HTTPSSERVER6: service="https"; break;
+#endif
+#ifdef SUPPORT_FTP
+	case FTPSERVER4:
+	case FTPSERVER6: service="ftp"; break;
+#endif
+#ifdef SUPPORT_SMB
+	case SMBSERVER4:
+	case SMBSERVER6: service="smb"; break;
+#endif
+	default: service="unk";
+	}
+	buffer_puts(buffer_1,service);
+      }
       buffer_putnlflush(buffer_1);
     }
+    if (limit_to_lan) {
+      int passed;
+      /* if the -L option is given, only accept connections from local
+       * or reserved IP ranges */
+      if (byte_equal(ip,12,V4mappedprefix)) {
+	unsigned char* ip4=(unsigned char*)ip+12;
+	passed = ip4[0]==127 ||	/* 127.0.0.1/8 */
+	    ip4[0]==10 ||	/* RFC1918: 10.0.0.0/8 */
+	    (ip4[0]==192 && ip4[1]==168) ||	/* RFC1918: 192.168.0.0/16 */
+	    (ip4[0]==172 && ip4[1]>=16 && ip4[1]<=31) ||	/* RFC1918: 172.16.0.0/12 */
+	    (ip4[0]==169 && ip4[1]==254 && ip4[2]!=0 && ip4[2]!=255);	/* RFC 5735: 169.254.0.0/16 */
+      } else {	/* IPv6 */
+	unsigned char* ip6=(unsigned char*)ip;
+	passed = byte_equal(ip6,16,V6loopback) ||	/* ::1 */
+	  (ip6[0]==0xfc && (ip6[1]&0xfe)==0) ||	/* RF4193 ULA fc00::/7 */
+	  (ip6[0]==0xfe && (ip6[1]&0xfc)==0xc0) ||	/* deprecated site-local */
+	  (ip6[0]==0xfe && (ip6[1]&0xfc)==0x80);	/* RFC5735 link-local */
+      }
+      if (!passed) {
+	timeout_secs=0;
+	punk=1;
+	if (logging) {
+	  buffer_puts(buffer_1,"close/outside_lan_drop ");
+	  buffer_putulong(buffer_1,n);
+	  buffer_putnlflush(buffer_1);
+	}
+      }
+    }
     if (punk && !timeout_secs) {
       io_close(n);
       continue;
@@ -594,7 +655,7 @@ static void accept_server_connection(int64 i,struct http_data* H,unsigned long f
 #warning TCP_NODELAY not defined
 #endif
 
-    if (io_fd(n)) {
+    if (io_fd_canwrite(n)) {
       struct http_data* h;
       h=(struct http_data*)malloc(sizeof(struct http_data));
       if (h) {
@@ -646,7 +707,12 @@ static void accept_server_connection(int64 i,struct http_data* H,unsigned long f
 #else
 	  fchdir(origdir);
 #endif
-	  if (init_serverside_tls(&h->ssl,n)) {
+#ifdef USE_OPENSSL
+	  if (init_serverside_tls(&h->ssl,n))
+#elif defined(USE_POLARSSL)
+	  if (init_serverside_tls(&h->ssl,&h->ssn,n))
+#endif
+	  {
 	    if (logging) {
 	      char a[FMT_ULONG];
 	      a[fmt_ulong(a,n)]=0;
@@ -655,7 +721,8 @@ static void accept_server_connection(int64 i,struct http_data* H,unsigned long f
 	    cleanup(n);
 	    continue;
 	  }
-	  changestate(h,HTTPSACCEPT);
+
+	  changestate(h,HTTPSACCEPT_CHECK);
 //	  h->t=HTTPSACCEPT;
 	  if (timeout_secs)
 	    io_timeout(n,nextftp);
@@ -709,9 +776,11 @@ static void accept_server_connection(int64 i,struct http_data* H,unsigned long f
 }
 
 #ifdef SUPPORT_HTTPS
+
 int handle_ssl_error_code(int sock,int code,int reading) {
 //  printf("handle_ssl_error_code(sock %d,code %d,reading %d)\n",sock,code,reading);
   switch (code) {
+#ifdef USE_OPENSSL
   case SSL_ERROR_WANT_READ:
     io_wantread(sock);
     io_dontwantwrite(sock);
@@ -720,18 +789,26 @@ int handle_ssl_error_code(int sock,int code,int reading) {
     io_wantwrite(sock);
     io_dontwantread(sock);
     return 0;
+#elif defined(USE_POLARSSL)
+  case POLARSSL_ERR_NET_WANT_READ:
+    io_wantread(sock);
+    io_dontwantwrite(sock);
+    return 0;
+  case POLARSSL_ERR_NET_WANT_WRITE:
+    io_wantwrite(sock);
+    io_dontwantread(sock);
+    return 0;
+#endif
+#ifdef USE_OPENSSL
   case SSL_ERROR_SYSCALL:
-    if (logging) {
-      int olderrno=errno;
-      buffer_puts(buffer_1,"io_error ");
-      buffer_putulong(buffer_1,sock);
-      buffer_putspace(buffer_1);
-      buffer_puterror(buffer_1);
-      buffer_puts(buffer_1,"\nclose/readerr ");
-      buffer_putulong(buffer_1,sock);
-      buffer_putnlflush(buffer_1);
-      errno=olderrno;
-    }
+#elif defined(USE_POLARSSL)
+  case POLARSSL_ERR_NET_RECV_FAILED:
+  case POLARSSL_ERR_NET_SEND_FAILED:
+  case POLARSSL_ERR_NET_CONN_RESET:
+    errno=ECONNRESET;
+#endif
+    // we already signal the error up and upsteam will then write an
+    // error message
     return -1;
   default:
     if (logging) {
@@ -741,14 +818,63 @@ int handle_ssl_error_code(int sock,int code,int reading) {
       buffer_putulong(buffer_1,sock);
       buffer_putnlflush(buffer_1);
     }
+    errno=EPROTO;
     return -1;
   }
 }
 
 void do_sslaccept(int sock,struct http_data* h,int reading) {
-  int r=SSL_get_error(h->ssl,SSL_accept(h->ssl));
+  int r;
+  if (h->t == HTTPSACCEPT_CHECK) {
+    unsigned char buf[10];
+    changestate(h,HTTPSACCEPT);
+    /* now, try to peek into the buffer, and see if it looks like an SSL
+     * handshake */
+    r=recv(sock,buf,sizeof(buf),MSG_PEEK);
+    if (r>5) {
+      /* Apparently the packets look radically different depending on
+       * whether it's TLS or SSLv2.  GREAT! */
+      /* first packet must be handshake */
+      if (buf[0]!=0x16 ||	/* content type: handshake */
+	  buf[1]>3 ||	/* version major */
+	  buf[2]>3 ||	/* version minor */
+	  buf[3]>1) {	/* length > 0x100 */
+	if (buf[0]!=0x80 ||
+	    buf[2]!=1 ||	/* Client Hello */
+	    buf[3]>3 ||	/* version major */
+	    buf[4]>3) {	/* version minor */
+	  tai6464 tarpit;
+	  /* this does not look like an SSL packet. */
+	  if (logging) {
+	    char tmp[100];
+	    buffer_puts(buffer_1,"close/not_ssl_traffic ");
+	    buffer_putulong(buffer_1,sock);
+	    buffer_putspace(buffer_1);
+	    buffer_put(buffer_1,tmp,fmt_ip6c(tmp,h->peerip));
+	    buffer_putnlflush(buffer_1);
+	  }
+	  io_dontwantread(sock);
+	  io_dontwantwrite(sock);
+	  --https_connections;
+	  changestate(h,PUNISHMENT);
+	  tarpit=now;
+	  tarpit.sec.x+=10;
+	  io_timeout(sock,tarpit);
+	  return;
+	}
+      }
+    }
+  }
+
+#ifdef USE_OPENSSL
+  r=SSL_get_error(h->ssl,SSL_accept(h->ssl));
 //  printf("do_sslaccept -> %d\n",r);
-  if (r==SSL_ERROR_NONE) {
+  if (r==SSL_ERROR_NONE)
+#elif defined(USE_POLARSSL)
+  r=ssl_handshake(&h->ssl);
+  if (r==0)
+#endif
+  {
 #if 0
     h->writefail=1;
 #endif
@@ -766,7 +892,7 @@ void do_sslaccept(int sock,struct http_data* h,int reading) {
 }
 #endif
 
-static void handle_read_misc(int64 i,struct http_data* H,unsigned long ftptimeout_secs,tai6464 nextftp) {
+static void handle_read_misc(int64 i,struct http_data* h,unsigned long ftptimeout_secs,tai6464 nextftp) {
   /* This is a TCP client connection waiting for input, i.e.
     *   - an HTTP connection waiting for a HTTP request, or
     *   - an FTP connection waiting for a command, or
@@ -775,21 +901,42 @@ static void handle_read_misc(int64 i,struct http_data* H,unsigned long ftptimeou
   char buf[8192];
   int l;
 #ifdef SUPPORT_HTTPS
-  assert(H->t != HTTPSRESPONSE);
-  if (H->t == HTTPSREQUEST) {
-    l=SSL_read(H->ssl,buf,sizeof(buf));
+  assert(h->t != HTTPSRESPONSE);
+  if (h->t == HTTPSREQUEST) {
+#ifdef USE_OPENSSL
+    l=SSL_read(h->ssl,buf,sizeof(buf));
+#elif defined(USE_POLARSSL)
+    l=ssl_read(&h->ssl,(unsigned char*)buf,sizeof(buf));
+#else
+#error fixme
+#endif
 //    printf("SSL_read(sock %d,buf %p,n %d) -> %d\n",i,buf,sizeof(buf),l);
+#ifdef USE_OPENSSL
     if (l==-1) {
-      l=SSL_get_error(H->ssl,l);
-//      printf("  error %d %s\n",l,ERR_error_string(l,0));
+      l=SSL_get_error(h->ssl,l);
       if (l==SSL_ERROR_WANT_READ || l==SSL_ERROR_WANT_WRITE) {
+#elif defined(USE_POLARSSL)
+    if (l<0) {
+      if (l==POLARSSL_ERR_NET_WANT_READ || l==POLARSSL_ERR_NET_WANT_WRITE) {
+#else
+#error fixme
+#endif
+//      printf("  error %d %s\n",l,ERR_error_string(l,0));
 	io_eagain(i);
 	if (handle_ssl_error_code(i,l,1)==-1) {
 	  cleanup(i);
 	  return;
 	}
 	l=-1;
-      } else l=-3;
+#ifdef USE_POLARSSL
+      } else if (l==POLARSSL_ERR_NET_RECV_FAILED) {
+	l=0;
+#endif
+      } else {
+	printf("got polarssl error %x\n",l);
+	errno=ECONNRESET;
+	l=-3;
+      }
     }
   } else
 #endif
@@ -811,20 +958,20 @@ ioerror:
       buffer_putnlflush(buffer_1);
     }
 #ifdef SUPPORT_FTP
-    if (H->t==FTPSLAVE) {
+    if (h->t==FTPSLAVE) {
       /* This is an FTP upload, it just finished. */
-      struct http_data* b=io_getcookie(H->buddy);
+      struct http_data* b=io_getcookie(h->buddy);
       assert(b);
       b->buddy=-1;
       iob_reset(&b->iob);
       iob_adds(&b->iob,"226 Got it.\r\n");
-      io_dontwantread(H->buddy);
-      io_wantwrite(H->buddy);
+      io_dontwantread(h->buddy);
+      io_wantwrite(h->buddy);
       if (chmoduploads)
-	fchmod(H->filefd,0644);
+	fchmod(h->filefd,0644);
       if (logging) {
 	struct stat ss;
-	if (fstat(H->filefd,&ss)==0) {
+	if (fstat(h->filefd,&ss)==0) {
 	  char a[FMT_ULONG];
 	  char b[FMT_ULONG];
 	  a[fmt_ulong(a,i)]=0;
@@ -837,10 +984,10 @@ ioerror:
     cleanup(i);
   } else if (l>0) {
     /* successfully read some data (l bytes) */
-    H->received+=l;
+    h->received+=l;
     tin1+=l;
 #ifdef SUPPORT_FTP
-    if (H->t==FTPCONTROL4 || H->t==FTPCONTROL6) {
+    if (h->t==FTPCONTROL4 || h->t==FTPCONTROL6) {
       if (ftptimeout_secs)
 	io_timeout(i,nextftp);
     } else {
@@ -848,39 +995,41 @@ ioerror:
 	io_timeout(i,next);
     }
 
-    if (H->t==FTPSLAVE) {
+    if (h->t==FTPSLAVE) {
       /* receive an upload */
       if (ftptimeout_secs)
-	io_timeout(H->buddy,nextftp);
-      if (write(H->filefd,buf,l)!=l)
+	io_timeout(h->buddy,nextftp);
+      if (write(h->filefd,buf,l)!=l)
 	goto ioerror;
     } else
 #endif
     {
       /* received a request */
-      array_catb(&H->r,buf,l);
-      if (array_failed(&H->r)) {
-	httperror(H,"500 Server Error","request too long.",0);
+//      printf("previous size: %lu, adding %lu bytes.\n",(unsigned long)array_bytes(&h->r),(unsigned long)l);
+      array_catb(&h->r,buf,l);
+      if (array_failed(&h->r)) {
+	httperror(h,"500 Server Error","request too long.",0);
 emerge:
 	io_dontwantread(i);
 	io_wantwrite(i);
-      } else if (array_bytes(&H->r)>MAX_HEADER_SIZE) {
-	httperror(H,"500 request too long","You sent too much header data",0);
-	array_reset(&H->r);
+      } else if (array_bytes(&h->r)>MAX_HEADER_SIZE) {
+	httperror(h,"500 request too long","You sent too much header data",0);
+	array_reset(&h->r);
 	goto emerge;
-      } else if ((l=header_complete(H))) {
+      } else if ((l=header_complete(h,i))) {
+	/* at this point we saw a \r\n\r\n */
 	long alen;
 pipeline:
-	/* The H->mimetype reference is here so that we don't count both the HTTP
+	/* The h->mimetype reference is here so that we don't count both the HTTP
 	 * connection and the first request on it as a dos attack. */
-	if (H->mimetype && new_request_from_ip(H->peerip,now.sec.x-4611686018427387914ULL)==1) {
-	  changestate(H,PUNISHMENT);
-//	  H->t=PUNISHMENT;
+	if (h->mimetype && new_request_from_ip(h->peerip,now.sec.x-4611686018427387914ULL)==1) {
+	  changestate(h,PUNISHMENT);
+//	  h->t=PUNISHMENT;
 	  if (logging) {
 	    char buf[IP6_FMT];
 	    char n[FMT_LONG];
 	    n[fmt_ulong(n,i)]=0;
-	    buf[fmt_ip6c(buf,H->peerip)]=0;
+	    buf[fmt_ip6c(buf,h->peerip)]=0;
 
 	    buffer_putmflush(buffer_1,"dos_tarpit2 ",n," ",buf,"\n");
 	  }
@@ -888,23 +1037,22 @@ pipeline:
 	    io_dontwantread(i);
 	    io_dontwantwrite(i);
 	  } else
-	    /* if we don't have timeouts enabled, just drop the
-	     * connection */
+	    /* if we don't have timeouts enabled, just drop the connection */
 	    cleanup(i);
 	  return;
 	}
 #ifdef SUPPORT_HTTPS
-	if (H->t==HTTPREQUEST || H->t==HTTPSREQUEST) {
-	  httpresponse(H,i,l);
-	  if (H->t == HTTPSREQUEST) changestate(H,HTTPSRESPONSE); // H->t=HTTPSRESPONSE;
+	if (h->t==HTTPREQUEST || h->t==HTTPSREQUEST) {
+	  httpresponse(h,i,l);
+	  if (h->t == HTTPSREQUEST) changestate(h,HTTPSRESPONSE); // h->t=HTTPSRESPONSE;
 	}
 #else
-	if (H->t==HTTPREQUEST)
-	  httpresponse(H,i,l);
+	if (h->t==HTTPREQUEST)
+	  httpresponse(h,i,l);
 #endif
 #ifdef SUPPORT_SMB
-	else if (H->t==SMBREQUEST) {
-	  if (smbresponse(H,i)==-1) {
+	else if (h->t==SMBREQUEST) {
+	  if (smbresponse(h,i)==-1) {
 	    cleanup(i);
 	    return;
 	  }
@@ -912,33 +1060,40 @@ pipeline:
 #endif
 #ifdef SUPPORT_FTP
 	else
-	  ftpresponse(H,i);
+	  ftpresponse(h,i);
 #endif
 #ifdef SUPPORT_PROXY
-	if (H->t != HTTPPOST
+	if (h->t != HTTPPOST
 #ifdef SUPPORT_HTTPS
-	 && H->t != HTTPSPOST
+	 && h->t != HTTPSPOST
 #endif
 	                     ) {
 #endif
-	  if (l < (alen=array_bytes(&H->r))) {
-	    char* c=array_start(&H->r);
+	  if (l < (alen=array_bytes(&h->r))) {
+	    char* c=array_start(&h->r);
+
+#if 0
+	    printf("continuation: consumed %lu bytes from buffer (%lu still there)\n",
+		   (unsigned long)l,
+		   (unsigned long)(alen-l));
+#endif
 	    byte_copy(c,alen-l,c+l);
-	    array_truncate(&H->r,1,alen-l);
-	    l=header_complete(H);
+	    array_truncate(&h->r,1,alen-l);
+	    l=header_complete(h,i);
 
 #if 0
 	    write(1,"\n\n",2);
-	    write(1,array_start(&H->r),array_bytes(&H->r));
+	    write(1,array_start(&h->r),array_bytes(&h->r));
 	    write(1,"\n\n",2);
 #endif
 
 	    if (l) {
-	      if (H->r.initialized) --H->r.initialized;
+	      /* this breaks SMB: */
+//	      if (h->r.initialized) --h->r.initialized;
 	      goto pipeline;
 	    }
 	  } else
-	    array_reset(&H->r);
+	    array_reset(&h->r);
 #ifdef SUPPORT_PROXY
 	}
 #endif
@@ -952,19 +1107,29 @@ int64 https_write_callback(int64 sock,const void* buf,uint64 n) {
   int l;
   struct http_data* H=io_getcookie(sock);
   if (!H) return -3;
+
 #if 0
   H->writefail=!H->writefail;
   if (H->writefail) { errno=EAGAIN; return -1; }
 #endif
   if (n>65536) n=65536;
+#ifdef USE_OPENSSL
   l=SSL_write(H->ssl,buf,n);
   if (l<0) {
     l=SSL_get_error(H->ssl,l);
+#elif defined(USE_POLARSSL)
+  l=ssl_write(&H->ssl,buf,n);
+  if (l<0) {
+#endif
     if (handle_ssl_error_code(sock,l,0)==-1) {
-      cleanup(sock);
+//      cleanup(sock);
       return -3;
     }
+#ifdef USE_OPENSSL
     if (l==SSL_ERROR_WANT_READ || l==SSL_ERROR_WANT_WRITE) {
+#elif defined(USE_POLARSSL)
+    if (l==POLARSSL_ERR_NET_WANT_READ || l==POLARSSL_ERR_NET_WANT_WRITE) {
+#endif
       l=-1; errno=EAGAIN;
     } else
       l=-3;
@@ -1011,6 +1176,7 @@ static void handle_write_misc(int64 i,struct http_data* h,uint64 prefetchquantum
 #endif
       cleanup(i);
     } else {	/* returned 0, i.e. we wrote it all */
+wroteitall:
 #ifdef SUPPORT_HTTPS
       if (h->t == HTTPSRESPONSE) h->t = HTTPSREQUEST;
 #endif
@@ -1058,7 +1224,7 @@ static void handle_write_misc(int64 i,struct http_data* h,uint64 prefetchquantum
 	  struct http_data* b=io_getcookie(h->buddy);
 	  if (b) {
 	    b->buddy=-1;
-	    iob_reset(&b->iob);
+//	    iob_reset(&b->iob);
 	    iob_adds(&b->iob,"226 Done.\r\n");
 	    io_dontwantread(h->buddy);
 	    io_wantwrite(h->buddy);
@@ -1080,6 +1246,8 @@ static void handle_write_misc(int64 i,struct http_data* h,uint64 prefetchquantum
 	h->prefetched_until+=2*prefetchquantum;
       }
     }
+    if (iob_bytesleft(&h->iob)==0)	/* optimization, not strictly necessary */
+      goto wroteitall;
   }
 }
 
@@ -1148,11 +1316,21 @@ int main(int argc,char* argv[],char* envp[]) {
   char* new_uid=0;
   char* chroot_to=0;
   unsigned long long prefetchquantum=0;
+#ifdef SUPPORT_MULTIPROC
   pid_t* Instances;
+#endif
+
+#if defined(DEBUG_EVENTS) || defined(MOREDEBUG) || defined(SMDEBUG) || defined(STATE_DEBUG)
+#ifdef __dietlibc__
+  fflush(stdout);
+#endif
+#endif
 
 #ifdef SUPPORT_HTTPS
+#ifdef USE_OPENSSL
   SSL_load_error_strings();
 #endif
+#endif
 
 #if defined(SUPPORT_CGI) || defined(SUPPORT_PROXY)
   _envp=envp;
@@ -1166,7 +1344,7 @@ int main(int argc,char* argv[],char* envp[]) {
 
     found=0;
     for (;;) {
-      int c=getopt(_argc,_argv,"HP:hnfFi:p:vVdDtT:c:u:Uaw:sSO:C:leEr:o:N:m:A:X:I:");
+      int c=getopt(_argc,_argv,"HP:hnfFi:p:vVdDtT:c:u:Uaw:sSO:C:lLeEr:o:N:m:A:X:I:");
       if (c==-1) break;
       switch (c) {
       case 'c':
@@ -1208,9 +1386,7 @@ int main(int argc,char* argv[],char* envp[]) {
 	  buffer fsb;
 #ifndef __MINGW32__
 	  if (chroot_to) { chdir(chroot_to); chroot(chroot_to); }
-	  if (new_uid) {
-	    prepare_switch_uid(new_uid);
-	  }
+	  prepare_switch_uid(new_uid);
 #endif
 	  if (!io_readfile(&savedir,".")) panic("open()");
 	  buffer_init(&fsb,(void*)read,forksock[1],fsbuf,sizeof fsbuf);
@@ -1297,15 +1473,25 @@ int main(int argc,char* argv[],char* envp[]) {
   }
 
   if (!geteuid()) {
-    struct rlimit rl;
+    struct rlimit orig,rl;
     long l;
+    getrlimit(RLIMIT_NOFILE,&orig);
 #ifdef RLIMIT_NPROC
     rl.rlim_cur=RLIM_INFINITY; rl.rlim_max=RLIM_INFINITY;
     setrlimit(RLIMIT_NPROC,&rl);
 #endif
-    for (l=0; l<20000; l+=500) {
+    for (l=orig.rlim_max; l<20000; l+=500) {
       rl.rlim_cur=l; rl.rlim_max=l;
-      if (setrlimit(RLIMIT_NOFILE,&rl)==-1) break;
+      if (setrlimit(RLIMIT_NOFILE,&rl)==-1) {
+	if (errno==EPERM) {
+	  /* We run as root but we still get EPERM?  That only happens
+	   * on linux-vserver.  There is no good way to handle this, so
+	   * we'll set the soft limit to the hard limit */
+	  orig.rlim_cur=orig.rlim_max;
+	  setrlimit(RLIMIT_NOFILE,&orig);
+	}
+	break;
+      }
     }
   }
 #endif
@@ -1324,9 +1510,12 @@ int main(int argc,char* argv[],char* envp[]) {
 
   for (;;) {
     int i;
-    int c=getopt(argc,argv,"HP:hnfFi:p:vVdDtT:c:u:Uaw:sSO:C:leEr:o:N:m:A:X:I:");
+    int c=getopt(argc,argv,"HP:hnfFi:p:vVdDtT:c:u:Uaw:sSO:C:lLeEr:o:N:m:A:X:I:");
     if (c==-1) break;
     switch (c) {
+    case 'L':
+      limit_to_lan=1;
+      break;
     case 'U':
       nouploads=1;
       break;
@@ -1533,6 +1722,7 @@ usage:
 #ifdef SUPPORT_FALLBACK_REDIR
 		  "\t-r url\tinstead of a 404, generate a redirect to url+localpart\n"
 #endif
+		  "\t-L\tonly accept connections from localhost or link/site local reserved IP addresses\n"
 		  );
       return 0;
     }
@@ -1553,6 +1743,7 @@ usage:
 #endif
     wglen=str_len(workgroup);
     wglen16=sizeof(workgroup_utf16)-Y;
+    iconv_close(i);
 
   }
   {
@@ -1722,6 +1913,7 @@ usage:
   }
   if (new_uid && switch_uid()==-1)
     panic("switch_uid");
+
 #endif
 
 #ifdef __MINGW32__
@@ -1775,8 +1967,10 @@ usage:
 
   } else {
 #endif
+#ifdef SUPPORT_MULTIPROC
     Instances=0;
     instances=0;
+#endif
 
 #ifdef __broken_itojun_v6__
     prepare_listen(s,&ct);
@@ -1838,6 +2032,17 @@ usage:
 
   connections=1;
 
+#if !defined(__OPTIMIZE__) && defined(__linux__)
+  /* make sure we can dump core even if we switched uid */
+  prctl(PR_SET_DUMPABLE,1,0,0);
+  /* and make sure we don't have a core dump size limit in place */
+  {
+    struct rlimit orig;
+    orig.rlim_cur=orig.rlim_max=RLIM_INFINITY;
+    setrlimit(RLIMIT_CORE,&orig);
+  }
+#endif
+
   for (;;) {
     int events;		/* accept new connections asap */
     int64 i;
@@ -1899,7 +2104,7 @@ usage:
 	while ((i=io_timeouted())!=-1) {
 	  struct http_data* x;
 #if defined(SUPPORT_HTTPS) && !defined(SUPPORT_MULTIPROC)
-	  if (ssh_timeout && (x=io_getcookie(i)) && x->t == HTTPSACCEPT) {
+	  if (ssh_timeout && (x=io_getcookie(i)) && x->t == HTTPSACCEPT_CHECK) {
 	    if (logging) {
 	      char numbuf[FMT_ULONG];
 	      numbuf[fmt_ulong(numbuf,i)]=0;
@@ -2019,7 +2224,7 @@ usage:
       else
 #endif
 #ifdef SUPPORT_HTTPS
-      if (H->t==HTTPSACCEPT)
+      if (H->t==HTTPSACCEPT || H->t==HTTPSACCEPT_CHECK)
 	do_sslaccept(i,H,1);
       else
 #endif
@@ -2033,6 +2238,8 @@ usage:
       }
       else {
 #ifdef SUPPORT_HTTPS
+	/* This looks wrong but isn't.  SSL has renegotiation.
+	   If that happens, we can be waiting for read when we are actually trying to write something */
 	if (H->t == HTTPSRESPONSE)
 	  handle_write_misc(i,H,prefetchquantum);
 	else
@@ -2096,16 +2303,19 @@ usage:
       if (h->t==FTPCONTROL4 || h->t==FTPCONTROL6) {
 	if (ftptimeout_secs)
 	  io_timeout(i,nextftp);
-      } else if (timeout_secs) {
+      } else
+#endif
+      if (timeout_secs) {
 	io_timeout(i,next);
-	if (h->t==FTPSLAVE) {
-	  io_timeout(h->buddy,nextftp);
+	if (h->buddy>0) {
+#ifdef SUPPORT_FTP
+	  if (h->t==FTPSLAVE) {
+	    io_timeout(h->buddy,nextftp);
+	  } else
+#endif
+	    io_timeout(h->buddy,next);
 	}
       }
-#else
-      if (timeout_secs)
-	io_timeout(i,next);
-#endif
 
 #ifdef SUPPORT_PROXY
       if (h->t==PROXYSLAVE)
@@ -2143,6 +2353,19 @@ usage:
   }
 #endif
   io_finishandshutdown();
+#ifdef SUPPORT_HTTPS
+#ifdef USE_OPENSSL
+  ERR_free_strings();
+#endif
+#endif
+#ifdef SUPPORT_SMB
+  {
+    extern iconv_t wc2utf8;
+    extern iconv_t utf82wc2;
+    iconv_close(wc2utf8);
+    iconv_close(utf82wc2);
+  }
+#endif
   _exit(0);
   return 0;
 }
