@@ -22,6 +22,12 @@
 #include <dirent.h>
 #ifdef __dietlibc__
 #include <md5.h>
+#elif defined(USE_POLARSSL)
+#include <polarssl/md5.h>
+#define MD5_CTX md5_context
+#define MD5Init md5_starts
+#define MD5Update md5_update
+#define MD5Final(out,ctx) md5_finish(ctx,out)
 #else
 #include <openssl/md5.h>
 #define MD5Init MD5_Init
@@ -60,7 +66,7 @@ char* http_header(struct http_data* r,char* h) {
 }
 
 static inline int issafe(unsigned char c) {
-  return (c!='"' && c!='%' && c>=' ' && c!='+' && c!=':' && c!='#');
+  return (c!='"' && c!='%' && (c>=' ' && c<0x7f) && c!='+' && c!=':' && c!='#');
 }
 
 size_t fmt_urlencoded(char* dest,const char* src,size_t len) {
@@ -93,6 +99,47 @@ void cathtml(array* a,char* s) {
   array_catb(a,buf,fmt_html(buf,s,len));
 }
 
+void cathtmlutf8(array* a,char* s) {
+  /* The purpose of this function is to convert a file name into UTF-8
+   * and escape HTML-relevant characters such as '<' and '&'. Chars that
+   * are not valid UTF-8 are assumed to be latin1 and converted */
+  size_t i,l,r;
+  char* buf;
+  r=0;
+  /* This will be a short string, a file name, so assuming all chars are
+   * '&', the max expansion is '&amp;', i.e. *5. */
+  l=strlen(s);
+  buf=alloca(l*5);
+  for (i=0; i<l; ++i) {
+    if (s[i]&0x80) {
+      size_t n=scan_utf8(s+i,l-i,NULL);
+      if (n==0) {
+	r+=fmt_utf8(buf+r,(unsigned char)(s[i]));
+      } else {
+	memcpy(buf+r,s+i,n);
+	i+=n-1;
+	r+=n;
+      }
+    } else {
+      const char* x=0;
+      size_t n;
+      switch (s[i]) {
+      case '&': x="&amp;"; n=5; break;
+      case '<': x="&lt;"; n=4; break;
+      case '>': x="&gt;"; n=4; break;
+      case '\n': x="<br>"; n=4; break;
+      }
+      if (x) {
+	memcpy(buf+r,x,n);
+	r+=n;
+      } else
+	buf[r++]=s[i];
+    }
+  }
+  array_catb(a,buf,r);
+}
+
+
 int http_dirlisting(struct http_data* h,DIR* D,const char* path,const char* arg) {
   long i,o,n;
   struct dirent* d;
@@ -145,6 +192,8 @@ int http_dirlisting(struct http_data* h,DIR* D,const char* path,const char* arg)
   array_cats(&c,path);
   {
     char* tmp=http_header(h,"User-Agent");
+    /* don't give wget the column sorting interface so wget -m does not
+     * mirror it needlessly */
     if (tmp && byte_equal(tmp,5,"Wget/"))
       array_cats(&c,"</h1>\n<table><tr><th>Name<th>Last Modified<th>Size\n");
     else {
@@ -173,7 +222,7 @@ int http_dirlisting(struct http_data* h,DIR* D,const char* path,const char* arg)
     catencoded(&c,base+ab[i].name);
     if (S_ISDIR(ab[i].ss.st_mode) || ab[i].todir) array_cats(&c,"/");
     array_cats(&c,"\">");
-    cathtml(&c,base+ab[i].name);
+    cathtmlutf8(&c,base+ab[i].name);
 #ifndef __MINGW32__
     if (S_ISLNK(ab[i].ss.st_mode)) array_cats(&c,"@"); else
 #endif
@@ -224,6 +273,8 @@ int add_proxy(const char* c) {
       x->proxyproto=FASTCGI;
     else if (c[0]=='S')
       x->proxyproto=SCGI;
+    else if (c[0]=='H')
+      x->proxyproto=HTTP;
     else
       goto nixgut;
     c+=2;
@@ -383,7 +434,7 @@ static size_t fmt_cgivars(char* dst,struct http_data* h,const char* uri,size_t u
       }
       while (*x!=':') {
 	char c=*x;
-	if (c>='a' && c<='z') c-='a'-'A';
+	if (c>='a' && c<='z') c-='a'-'A';	/* toupper */
 	if (c=='-') c='_'; else
 	if (c<'A' || c>'Z') {
 	  dst=olddst;
@@ -469,7 +520,9 @@ freeandfail:
 	if (x->proxyproto == SCGI) {
 	  size_t l=fmt_cgivars(0,ctx_for_sockfd,c,matches.rm_eo,dir,0);
 	  char* x,* y;
-	  if (!array_allocate(&ctx_for_gatewayfd->r,1,l+fmt_ulong(0,l)+3))
+	  /* array_allocate gets the index of the last element you want
+	   * to access, not the number of bytes; so +1, not +2 */
+	  if (!array_allocate(&ctx_for_gatewayfd->r,1,l+fmt_ulong(0,l)+1))
 	    goto freeandfail;
 	  x=array_start(&ctx_for_gatewayfd->r);
 	  x+=fmt_ulong(x,l);
@@ -576,7 +629,7 @@ freeandfail:
 	    }
 	  }
 	} else if (x->proxyproto==HTTP) {
-	  size_t size_of_header=header_complete(ctx_for_sockfd);
+	  size_t size_of_header=header_complete(ctx_for_sockfd,sockfd);
 	  size_t i;
 	  char* x=array_start(&ctx_for_sockfd->r);
 	  for (i=0; i<size_of_header && x[i]!='\n'; ++i)
@@ -604,6 +657,10 @@ freeandfail:
 	case 'P':
 		  if (tmp[1]=='O')
 		    method=(ctx_for_sockfd->t==HTTPREQUEST)?"POST":"POST/SSL";
+#ifdef SUPPORT_DAV
+		  else if (tmp[1]=='R')
+		    method=(ctx_for_sockfd->t==HTTPREQUEST)?"PROPFIND":"PROPFIND/SSL";
+#endif
 		  else
 		    method=(ctx_for_sockfd->t==HTTPREQUEST)?"PUT":"PUT/SSL";
 		  break;
@@ -612,7 +669,13 @@ freeandfail:
 	switch (*tmp) {
 	case 'H': method="HEAD"; break;
 	case 'G': method="GET"; break;
-	case 'P': method=(tmp[1]=='O')?"POST":"PUT"; break;
+	case 'P': method=(tmp[1]=='O')?"POST":
+#ifdef SUPPORT_DAV
+		  ((tmp[1]=='R')?"PROPFIND":"PUT");
+#else
+		  "PUT";
+#endif
+		  break;
 	}
 #endif
 	buffer_putm(buffer_1,method,x->port?"/PROXY ":"/CGI ");
@@ -635,7 +698,6 @@ freeandfail:
 
       if (x->port) {
 	/* proxy mode */
-	changestate(ctx_for_gatewayfd,PROXYSLAVE);
 	if (x->port>0xffff) {	/* unix domain socket mode */
 	  fd_to_gateway=socket(AF_UNIX,SOCK_STREAM,0);
 	} else
@@ -644,6 +706,7 @@ freeandfail:
 	ctx_for_gatewayfd->myfd=fd_to_gateway;
 #endif
 	if (fd_to_gateway==-1) goto punt2;
+	changestate(ctx_for_gatewayfd,PROXYSLAVE);
 	if (!io_fd(fd_to_gateway)) {
 punt:
 	  io_close(fd_to_gateway);
@@ -663,6 +726,7 @@ punt2:
 	    if (errno!=EINPROGRESS)
 	      goto punt;
 	}
+	io_fd_canwrite(fd_to_gateway);
 	if (logging) {
 	  char tmp[100];
 	  char bufsockfd[FMT_ULONG];
@@ -746,7 +810,7 @@ punt2:
 	    free(ctx_for_gatewayfd);
 	    return -1;
 	  }
-	  if (!io_fd(fd_to_gateway)) {
+	  if (!io_fd_canwrite(fd_to_gateway)) {
 	    httperror(ctx_for_sockfd,"502 Gateway Broken",c,*ctx_for_sockfd->r.p=='H'?1:0);
 	    io_close(fd_to_gateway);
 	    free(ctx_for_gatewayfd);
@@ -781,8 +845,8 @@ punt2:
       *   - the header and possibly some data left in ctx_for_sockfd->r.
       * Want:
       *   - leave the data (not the header) in ctx_for_sockfd->r.
-      *   - set ctl_for_gatewayfd->still_to_copy to Content-Length.
-      *   - set ctl_for_sockfd->still_to_copy to Content-Length -
+      *   - set ctx_for_gatewayfd->still_to_copy to Content-Length.
+      *   - set ctx_for_sockfd->still_to_copy to Content-Length -
       *     the size of the copied data.  If that is non-zero, set t to
       *     HTTPPOST.
       */
@@ -804,37 +868,51 @@ punt2:
 	    const char contmsg[]="HTTP/1.1 100 Continue\r\n\r\n";
 	    /* if this fails, tough luck.  I'm not bloating my state
 	      * engine for this crap. */
+#ifdef SUPPORT_HTTPS
+	    if (ctx_for_sockfd->t==HTTPSREQUEST)
+#if defined(USE_OPENSSL)
+	      SSL_write(ctx_for_sockfd->ssl,contmsg,sizeof(contmsg)-1);
+#elif defined(USE_POLARSSL)
+	      ssl_write(&ctx_for_sockfd->ssl,(const unsigned char*)contmsg,sizeof(contmsg)-1);
+#else
+#warn fixme update SSL code in http.c
+#endif
+	    else
+#endif
 	    io_trywrite(sockfd,contmsg,sizeof(contmsg)-1);
 	  }
 	}
 
 	/* figure out how much data we have */
 	{
-	  size_t size_of_header=header_complete(ctx_for_sockfd);
+	  size_t size_of_header=header_complete(ctx_for_sockfd,sockfd);
 	  size_t size_of_data_in_packet=array_bytes(&ctx_for_sockfd->r) - size_of_header - 1;
 	    /* the -1 is for the \0 we appended */
 
+//	  printf("proxy_connection: size_of_header=%lu, size_of_data_in_packet=%lu, content_length=%lu\n",size_of_header,size_of_data_in_packet,content_length);
+
 #ifdef SUPPORT_HTTPS
 	  if (ctx_for_sockfd->t==HTTPSREQUEST)
 	    changestate(ctx_for_sockfd,HTTPSPOST);
 	  if (ctx_for_sockfd->t!=HTTPSPOST)
 #endif
 	  changestate(ctx_for_sockfd,HTTPPOST);
+
+	  /* slight complication: we might have more data already than
+	   * we need for this request, if the content length is small
+	   * and the client uses pipelining and added the next request
+	   * already. */
+	  if (size_of_data_in_packet > content_length)
+	    size_of_data_in_packet = content_length;
+
 	  if (size_of_data_in_packet) {
 	    byte_copy(array_start(&ctx_for_sockfd->r),
 		      size_of_data_in_packet,
 		      array_start(&ctx_for_sockfd->r)+size_of_header);
 	    array_truncate(&ctx_for_sockfd->r,1,size_of_data_in_packet);
-/*
-	    if (content_length>size_of_data_in_packet)
-	      ctx_for_sockfd->still_to_copy-=size_of_data_in_packet;
-	    else
-	      ctx_for_sockfd->still_to_copy=0;
-*/
-	  } else {
+	  } else
 	    array_trunc(&ctx_for_sockfd->r);
-	    ctx_for_sockfd->still_to_copy=content_length;
-	  }
+	  ctx_for_sockfd->still_to_copy=content_length;
 
 	  if (ctx_for_gatewayfd->still_to_copy && array_bytes(&ctx_for_sockfd->r))
 	    io_wantwrite(fd_to_gateway);
@@ -846,6 +924,8 @@ punt2:
 	  else
 	    io_dontwantread(sockfd);
 
+//	  printf("proxy_connection: ctx_for_sockfd->still_to_copy=%lu, ctx_for_gatewayfd->still_to_copy=%lu\n",ctx_for_sockfd->still_to_copy, ctx_for_gatewayfd->still_to_copy);
+
 	}
       }
 
@@ -975,7 +1055,6 @@ nextpacket:
 	if (case_starts(buf,"Status:")) {
 	  --buf; ++i;
 	  memcpy(buf,"HTTP/1.1 ",9);
-	  ++i;
 	} else
 	  needheader=1;
       } else if (byte_diff(buf,5,"HTTP/"))
@@ -1029,10 +1108,16 @@ int read_http_post(int sockfd,struct http_data* H) {
   if (l>sizeof(buf)) l=sizeof(buf);
 #ifdef SUPPORT_HTTPS
   if (H->t==HTTPSPOST) {
+#ifdef USE_OPENSSL
     i=SSL_read(H->ssl,buf,l);
     if (i<0) {
       i=SSL_get_error(H->ssl,i);
       if (l==SSL_ERROR_WANT_READ || l==SSL_ERROR_WANT_WRITE) {
+#elif defined(USE_POLARSSL)
+    i=ssl_read(&H->ssl,(unsigned char*)buf,l);
+    if (i<0) {
+      if (l==POLARSSL_ERR_NET_WANT_READ || l==POLARSSL_ERR_NET_WANT_WRITE) {
+#endif
 	io_eagain(sockfd);
 	if (handle_ssl_error_code(sockfd,i,1)==-1)
 	  return -1;
@@ -1042,23 +1127,35 @@ int read_http_post(int sockfd,struct http_data* H) {
   } else
 #endif
   i=read(sockfd,buf,l);
-//  printf("read_http_post: want to read %ld bytes from %d; got %d\n",l,sockfd,i);
+#ifdef MOREDEBUG
+  printf("read_http_post: want to read %ld bytes from %d; got %d\n",(long)l,sockfd,i);
+#endif
   if (i<1) return -1;
 
   H->received+=i;
   H->still_to_copy-=i;
-//  printf("still_to_copy read_http_post: %p %llu -> %llu\n",H,H->still_to_copy+i,H->still_to_copy);
+#ifdef MOREDEBUG
+  printf("still_to_copy read_http_post: %p %llu -> %llu\n",H,H->still_to_copy+i,H->still_to_copy);
+#ifdef STATE_DEBUG
+  {
+    struct http_data* mybuddy=io_getcookie(H->buddy);
+    printf("read_http_post: my state is %s, my buddy's state is %s\n",state2string(H->t),state2string(mybuddy->t));
+  }
+#endif
+#endif
   /* we got some data.  Now, for FastCGI we need to add a header before
    * writing it to the proxy */
   if (H->proxyproto==FASTCGI) {
     char tmp[8]="\x01\x05\x00\x01\x00\x00\x00\x00";
     tmp[4]=i>>8;
     tmp[5]=i&0xff;
+
     array_catb(&H->r,tmp,8); /* FCGI_Record: FCGI_STDIN (5) */
   }
   array_catb(&H->r,buf,i);
-  if (H->proxyproto==FASTCGI && H->still_to_copy==0)
+  if (H->proxyproto==FASTCGI && H->still_to_copy==0) {
     array_catb(&H->r,"\x01\x05\x00\x01\x00\x00\x00\x00",8); /* FCGI_Record: FCGI_STDIN (5) */
+  }
   if (array_failed(&H->r))
     return -1;
   return 0;
@@ -1176,6 +1273,51 @@ int http_redirect(struct http_data* h,const char* Filename) {
   return 0;
 }
 
+#ifdef SUPPORT_DIR_REDIRECT
+void do_dir_redirect(struct http_data* h,const char* filename,int64 s) {
+  char* nh;
+  int i;
+  char* host=http_header(h,"Host");
+#ifdef SUPPORT_HTTPS
+  const char* proto=h->t==HTTPSREQUEST?"https://":"http://";
+#else
+  const char* proto="http://";
+#endif
+  size_t hl;
+  if (!host) return;
+  hl=str_chr(host,'\n');
+  if (hl && host[hl-1]=='\r') --hl;
+  nh=malloc((strlen(filename)+hl)*2+300);
+  if (!nh) {
+    if (logging) {
+      char numbuf[FMT_ULONG];
+      numbuf[fmt_ulong(numbuf,s)]=0;
+      buffer_putmflush(buffer_1,"outofmemory ",numbuf,"\n");
+    }
+    cleanup(s);
+    return;
+  }
+  i=fmt_str(nh,"HTTP/1.0 302 Over There\r\nServer: " RELEASE "\r\nLocation: ");
+  i+=fmt_str(nh+i,proto);
+  i+=fmt_strn(nh+i,host,hl);
+  i+=fmt_str(nh+i,filename);
+  i+=fmt_str(nh+i,"/\r\nContent-Type: text/html\r\nContent-Length: ");
+  i+=fmt_ulong(nh+i,strlen(filename)+hl+23);
+  i+=fmt_str(nh+i,"\r\n\r\n");
+  i+=fmt_str(nh+i,"Look <a href=");
+  i+=fmt_str(nh+i,proto);
+  i+=fmt_strn(nh+i,host,hl);
+  i+=fmt_str(nh+i,filename);
+  i+=fmt_str(nh+i,"/>here!</a>\n");
+  if (logging) {
+    char numbuf[FMT_ULONG];
+    numbuf[fmt_ulong(numbuf,s)]=0;
+    buffer_putmflush(buffer_1,"dir_redirect ",numbuf,"\n");
+  }
+  iob_addbuf_free(&h->iob,nh,strlen(nh));
+}
+#endif
+
 int64 http_openfile(struct http_data* h,char* filename,struct stat* ss,int sockfd,int nobody) {
 #ifdef SUPPORT_PROXY
   int noproxy=0;
@@ -1231,6 +1373,7 @@ int64 http_openfile(struct http_data* h,char* filename,struct stat* ss,int sockf
       Filename[i+1]=':';
   /* fourth, try to do some el-cheapo virtual hosting */
   if (!(s=http_header(h,"Host"))) {
+makefakeheader:
     /* construct artificial Host header from IP */
     s=alloca(IP6_FMT+7);
     i=fmt_ip6c(s,h->myip);
@@ -1238,6 +1381,14 @@ int64 http_openfile(struct http_data* h,char* filename,struct stat* ss,int sockf
     i+=fmt_ulong(s+i,h->myport);
     s[i]=0;
   } else {
+    size_t k;
+    char* t;
+    for (k=0; s[k] && s[k]!='/' && s[k]>' '; ++k) ;
+    t=alloca(k+2);
+    memcpy(t,s,k);
+    t[k]='\r'; t[k+1]=0;
+    s=t;
+    if (s[0]=='.' || !s[0]) goto makefakeheader;
     if (virtual_hosts>=0) {
       char* tmp;
       int j=str_chr(s,'\r');
@@ -1454,6 +1605,15 @@ itsafile:
       }
       return -1;
     }
+#ifdef SUPPORT_DIR_REDIRECT
+    if (S_ISDIR(ss->st_mode)) {
+      /* someone asked for http://example.com/foo
+       * when he should have asked for http://example.com/foo/
+       * redirect */
+      do_dir_redirect(h,Filename,sockfd);
+      return -4;
+    }
+#endif
     h->mimetype=mimetype(Filename,fd);
 foundcgi:
 #ifdef SUPPORT_PROXY
@@ -1709,6 +1869,9 @@ void httpresponse(struct http_data* h,int64 s,long headerlen) {
   int head,post;
 #ifdef SUPPORT_PUT
   int put;
+#endif
+#ifdef SUPPORT_DAV
+  int propfind;
 #endif
   char* c;
   const char* m;
@@ -1722,6 +1885,9 @@ void httpresponse(struct http_data* h,int64 s,long headerlen) {
   if (byte_diff(c,5,"GET /") && byte_diff(c,6,"POST /") &&
 #ifdef SUPPORT_PUT
       byte_diff(c,5,"PUT /") &&
+#endif
+#ifdef SUPPORT_DAV
+      byte_diff(c,10,"PROPFIND /") &&
 #endif
       byte_diff(c,6,"HEAD /")) {
 e400:
@@ -1742,6 +1908,10 @@ e400:
     post=c[1]=='O';
 #ifdef SUPPORT_PUT
     put=c[1]=='U';
+#endif
+#ifdef SUPPORT_DAV
+    if ((propfind=c[1]=='R'))
+      c+=5;	// we will advance by 4, so make sure it's 9 total
 #endif
     c+=(head||post)?5:4;
     for (d=c; *d!=' '&&*d!='\t'&&*d!='\n'&&*d!='\r'; ++d) ;
@@ -1846,7 +2016,7 @@ e404:
 	    buffer_putsflush(buffer_1,"\n");
 	  }
 
-	  c+=fmt_str(c,"HTTP/1.1 200 Here you go\r\nContent-Type: text/html\r\nConnection: ");
+	  c+=fmt_str(c,"HTTP/1.1 200 Here you go\r\nContent-Type: text/html; charset=utf-8\r\nConnection: ");
 	  c+=fmt_str(c,h->keepalive?"keep-alive":"close");
 	  c+=fmt_str(c,"\r\nServer: " RELEASE "\r\nContent-Length: ");
 	  c+=fmt_ulong(c,h->blen);
@@ -1894,7 +2064,7 @@ e404:
 	  if (byte_equal(c,6,"bytes=")) {
 	    size_t i;
 	    c+=6;
-	    n=scan_range(c,ranges,sizeof(ranges)/sizeof(ranges[0])/2,ss.st_size);
+	    n=scan_range(c,ranges,sizeof(ranges)/sizeof(ranges[0])/2,ss.st_size-1);
 
 	    /* the ranges could still be bogus, i.e. 4-2, or the sum
 	      * could be more than just sending the whole file. */
@@ -1977,11 +2147,12 @@ rangekaputt:
 	} else {
 	  c+=fmt_str(c,h->mimetype);
 	  c+=fmt_str(c,"\r\nContent-Length: ");
-	  c+=fmt_ulonglong(c,range_last-range_first);
+	  if (range_last==ss.st_size) --range_last;
+	  c+=fmt_ulonglong(c,range_last-range_first+1);
 	}
 
 	c+=fmt_str(c,"\r\nDate: ");
-	c+=fmt_httpdate(c,now.sec.x&0xfffffffffffull);
+	c+=fmt_httpdate(c,now.sec.x-4611686018427387914ULL);
 
 	c+=fmt_str(c,"\r\nLast-Modified: ");
 	c+=fmt_httpdate(c,ss.st_mtime);
@@ -1993,16 +2164,16 @@ rangekaputt:
 	  c+=fmt_str(c,"gzip");
 #endif
 	}
-	if ((range_first || range_last!=ss.st_size)) {
+	if (!head && (range_first || range_last!=ss.st_size)) {
 	  c+=fmt_str(c,"\r\nContent-Range: bytes ");
 	  c+=fmt_ulonglong(c,range_first);
 	  c+=fmt_str(c,"-");
-	  c+=fmt_ulonglong(c,range_last-1);
+	  c+=fmt_ulonglong(c,range_last);
 	  c+=fmt_str(c,"/");
 	  c+=fmt_ulonglong(c,ss.st_size);
 	}
 	if (range_first>ss.st_size) {
-	  free(c);
+	  free(h->hdrbuf);
 	  httperror(h,"416 Bad Range","The requested range can not be satisfied.",head);
 	  buffer_puts(buffer_1,"error_416 ");
 	} else {
@@ -2044,7 +2215,7 @@ rangekaputt:
 	      }
 	      iob_addbuf_free(&h->iob,buf,flen);
 	    } else
-	      iob_addfile_close(&h->iob,fd,range_first,range_last-range_first);
+	      iob_addfile_close(&h->iob,fd,range_first,range_last-range_first+1);
 	  } else
 	    if (fd!=-1) io_close(fd);
 	  if (logging) {
@@ -2083,6 +2254,7 @@ rangekaputt:
 	  buffer_putlogstr(buffer_1,(tmp=http_header(h,"Host"))?tmp:buf);
 	  buffer_putsflush(buffer_1,"\n");
 	}
+	h->filefd=-1;	/* iob_addfile_close closes the file for us, we don't want cleanup to close it again */
       }
     }
   }
@@ -2093,8 +2265,8 @@ fini:
 
 
 #ifdef SUPPORT_PROXY
-void handle_read_proxypost(int64 i,struct http_data* H) {
-  switch (proxy_is_readable(i,H)) {
+void handle_read_proxypost(int64 i,struct http_data* h) {
+  switch (proxy_is_readable(i,h)) {
   case -1:
     {
       if (logging) {
@@ -2109,15 +2281,19 @@ void handle_read_proxypost(int64 i,struct http_data* H) {
   }
 }
 
-void handle_read_httppost(int64 i,struct http_data* H) {
+void handle_read_httppost(int64 i,struct http_data* h) {
   /* read POST data. */
-//	printf("read POST data state for %d\n",i);
-  if (H->still_to_copy) {
-    if (array_bytes(&H->r)>0) {
-//	    printf("  but there was still data in H->r!\n");
+#ifdef MOREDEBUG
+	printf("read POST data state for %d\n",(int)i);
+#endif
+  if (h->still_to_copy) {
+    if (array_bytes(&h->r)>0) {
+#ifdef MOREDEBUG
+	    printf("  but there was still data in h->r!\n");
+#endif
       io_dontwantread(i);
-      io_wantwrite(H->buddy);
-    } else if (read_http_post(i,H)==-1) {
+      io_wantwrite(h->buddy);
+    } else if (read_http_post(i,h)==-1) {
       if (logging) {
 	char a[FMT_ULONG];
 	a[fmt_ulong(a,i)]=0;
@@ -2125,23 +2301,31 @@ void handle_read_httppost(int64 i,struct http_data* H) {
       }
       cleanup(i);
     } else {
-//	    printf("  read something\n");
+#ifdef MOREDEBUG
+	    printf("  read something\n");
+#endif
       io_dontwantread(i);
-      io_wantwrite(H->buddy);
+      io_wantwrite(h->buddy);
     }
   } else {
     /* should not happen */
     io_dontwantread(i);
-//	  printf("ARGH!!!\n");
+#ifdef MOREDEBUG
+	  printf("ARGH!!!\n");
+#endif
   }
 }
 
 void handle_write_proxypost(int64 i,struct http_data* h) {
   struct http_data* H=io_getcookie(h->buddy);
   /* do we have some POST data to write? */
-//	printf("event: write POST data (%llu) to proxy on %d\n",h->still_to_copy,i);
+#ifdef MOREDEBUG
+	printf("event: write POST data (%llu) to proxy on %d\n",h->still_to_copy,(int)i);
+#endif
   if (!array_bytes(&H->r)) {
-//	  printf("  but nothing here to write!\n");
+#ifdef MOREDEBUG
+	  printf("  but nothing here to write!\n");
+#endif
     io_dontwantwrite(i);	/* nope */
     io_wantread(h->buddy);
   } else {
@@ -2150,11 +2334,24 @@ void handle_write_proxypost(int64 i,struct http_data* h) {
       char* c=array_start(&H->r);
       long alen=array_bytes(&H->r);
       long l;
-//	    printf("%ld bytes still in H->r (%ld in h->r), still to copy: %lld (%lld in h)\n",alen,(long)array_bytes(&h->r),H->still_to_copy,h->still_to_copy);
-      if (alen>h->still_to_copy) alen=h->still_to_copy;
+#ifdef MOREDEBUG
+      printf("%ld bytes still in H->r (%ld in h->r), still to copy: %lld (%lld in h)\n",alen,(long)array_bytes(&h->r),H->still_to_copy,h->still_to_copy);
+#endif
+
+      if (h->proxyproto!=FASTCGI) {
+	/* this looks like the right thing to sanity-check but it is not
+	 * in the fastcgi case.  For fastcgi, we have to append an 8
+	 * byte header for each chunk, and if it's the last chunk, we
+	 * have to append another 8 byte header.  So alen can be 8 or 16
+	 * bytes off. */
+	if (alen>h->still_to_copy) alen=h->still_to_copy;
+      }
+
       if (alen==0) goto nothingmoretocopy;
       l=write(i,c,alen);
-//	    printf("wrote %ld bytes (wanted to write %ld; had %lld still to copy)\n",l,alen,H->still_to_copy);
+#ifdef MOREDEBUG
+	    printf("wrote %ld bytes (wanted to write %ld; had %lld still to copy)\n",l,alen,H->still_to_copy);
+#endif
       if (l<1) {
 	/* ARGH!  Proxy crashed! *groan* */
 	if (logging) {
@@ -2194,14 +2391,7 @@ nothingmoretocopy:
   }
 }
 
-void handle_write_httppost(int64 i,struct http_data* h) {
-  int64 r;
-#ifdef SUPPORT_HTTPS
-  if (h->t==HTTPSPOST)
-    r=iob_write(i,&h->iob,https_write_callback);
-  else 
-#endif
-  r=iob_send(i,&h->iob);
+static void handle_write_error(int64 i,struct http_data* h,int64 r) {
   if (r==-1)
     io_eagain(i);
   else if (r<=0) {
@@ -2237,6 +2427,41 @@ void handle_write_httppost(int64 i,struct http_data* h) {
     h->sent+=r;
 }
 
+void handle_write_httppost(int64 i,struct http_data* h) {
+  int64 r;
+#ifdef SUPPORT_HTTPS
+  if (h->t==HTTPSPOST)
+    r=iob_write(i,&h->iob,https_write_callback);
+  else 
+#endif
+  r=iob_send(i,&h->iob);
+  if (r > 0 && iob_bytesleft(&h->iob)==0) {
+    /* We wrote something and there is no more data left in the iob. */
+    /* Since we do not buffer the whole data from the proxy, there could
+     * be more data incoming from the proxy.  If this was the last
+     * batch, then the proxy connection has closed itself and set our
+     * buddy to -1. */
+    if (h->buddy==-1) {
+      if (logging) {
+	char a[FMT_ULONG];
+	char r[FMT_ULONG];
+	char s[FMT_ULONG];
+	a[fmt_ulong(a,i)]=0;
+	r[fmt_ulonglong(r,h->received)]=0;
+	s[fmt_ulonglong(s,h->sent)]=0;
+	buffer_putmflush(buffer_1,"close/proxydone ",a," ",r," ",s,"\n");
+      }
+      cleanup(i);
+    } else {
+      /* The proxy has more data for us */
+      io_dontwantwrite(i);
+      io_wantread(h->buddy);
+    }
+    return;
+  }
+  handle_write_error(i,h,r);
+}
+
 void handle_write_proxyslave(int64 i,struct http_data* h) {
   /* the connect() to the proxy just finished or failed */
   struct http_data* H;
@@ -2260,7 +2485,9 @@ kaputt:
   }
   /* it worked.  We wrote the header.  Now see if there is
     * POST data to write.  h->still_to_copy is Content-Length. */
-//	printf("wrote header to %d for %d; Content-Length: %d\n",(int)i,(int)h->buddy,(int)h->still_to_copy);
+#ifdef MOREDEBUG
+	printf("wrote header to %d for %d; Content-Length: %d\n",(int)i,(int)h->buddy,(int)h->still_to_copy);
+#endif
   changestate(h,PROXYPOST);
   array_trunc(&h->r);
   if (h->still_to_copy) {
@@ -2289,7 +2516,7 @@ kaputt:
 	H->r.initialized=0;
       }
     }
-    handle_write_httppost(i,H);
+    handle_write_error(i,H,iob_send(i,&H->iob));
   } else {
     io_dontwantwrite(i);
     io_wantread(i);
@@ -2429,6 +2656,7 @@ void forkslave(int fd,buffer* in,int savedir) {
       exit(127);
     }
     close(s);
+    --https_connections;
     return;
   }
 #endif
