@@ -19,6 +19,7 @@
 #include <sys/statvfs.h>
 #include <fnmatch.h>
 #include <dirent.h>
+#include <ctype.h>
 
 #include <stdio.h>
 
@@ -152,8 +153,40 @@ struct smbheader {
 };
 #endif
 
+static int globmatch_int(const char* pattern,const char* string,const char* ldot) {
+  while (*pattern) {
+    switch (*pattern) {
+    case '?':
+      if (!*string) return *pattern==0;
+      break;
+    case '<':
+      if (string>ldot)
+	return 0;
+    case '*':
+      /* pattern="*.x" and string=".x"? */
+      if (globmatch_int(pattern+1,string,ldot)) return 1;
+      /* pattern="*.x" and string="a.x"? */
+      if (*string==0) return 0;
+      ++string;
+      continue;
+    default:
+      if (tolower(*string) != tolower(*pattern)) return 0;
+      if (!*string) return 1;
+    }
+    ++pattern;
+    ++string;
+  }
+  return *string==0;
+}
+
+static int globmatch(const char* pattern,const char* string) {
+  const char* ldot=strrchr(string,'.');
+  if (!ldot) ldot=string+strlen(string);
+  return globmatch_int(pattern,string,ldot);
+}
+
 static int hasandx(unsigned char code) {
-  return !strchr("\x04\x72\x71\x2b\x32\x80\xa0",code);
+  return !strchr("\x04\x72\x71\x2b\x32\x80\xa0\x23",code);
 }
 
 static const size_t netbiosheadersize=4;
@@ -165,23 +198,26 @@ struct smb_response {
 };
 
 #ifdef DEBUG
-void hexdump(char* buf,size_t len) {
+static void hexdump(char* buf,size_t len) {
   size_t i,j;
-  char y[9];
-  y[8]=0;
-  printf("sending:\n");
+  char y[17];
+  y[16]=0;
+//  printf("sending:\n");
   for (i=j=0; i<len; ++i) {
+    if (j==16) j=0;
     y[j]=buf[i];
-    if (y[j]<' ') y[j]='.';
-    if (++j==8) j=0;
+    if (y[j]<' ' || y[j]==0x7f) y[j]='.';
+    ++j;
     printf("%02x",(unsigned char)(buf[i]));
-    if (i%8<7)
-      putchar(' ');
-    else
-      printf("   %s\n",y);
+    switch (i%16) {
+    case 15: printf("   %s\n",y); break;
+    case 7: putchar(' ');	// fallthrough
+    default: putchar(' '); break;
+    }
   }
   y[j]=0;
-  printf("%*s%s\n",(int)((9-j)*3-1),"",y);
+  if (j<16)
+    printf("%*s%s\n",(int)((17-j)*3)-1,"",y);
 }
 #endif
 
@@ -270,7 +306,7 @@ static int validate_smb_packet(unsigned char* pkt,unsigned long len) {
     {
       size_t bytecountofs=1+*x*2;
       size_t bytecount;
-      bytecount=uint16_read(x+bytecountofs);
+      bytecount=uint16_read((const char*)x+bytecountofs);
       if (bytecount>len || x+bytecountofs+2+bytecount>pkt+len) return -1;
     }
     if (!hasandx(pkt[4])) return 0;
@@ -282,7 +318,7 @@ static int validate_smb_packet(unsigned char* pkt,unsigned long len) {
 	return -1;
       /* we know that the byte count is within the packet */
       /* read it and check whether it's ok, too */
-      bytecount=uint16_read(x+1+*x*2);
+      bytecount=uint16_read((const char*)x+1+*x*2);
       if (!range_arrayinbuf(pkt,len,x+3+bytecount,*x,2))
 	return -1;
       if (x[1]==0xff) return 0;
@@ -347,8 +383,13 @@ static int smb_handle_negotiate_request(unsigned char* c,size_t len,struct smb_r
     "\x11"	// word count 17
     "xx"	// dialect index; ofs 1
     "\x02"	// security mode, for NT: plaintext passwords XOR unicode
+#if 0
     "\x02\x00"	// Max Mpx Count 2
     "\x01\x00"	// Max VCs 1
+#else
+    "\x10\x00"	// Max Mpx Count 2
+    "\x10\x00"	// Max VCs 1
+#endif
     "\x04\x41\x00\x00"	// Max Buffer Size (16644, like XP)
     "\x00\x00\x01\x00"	// Max Raw Buffer (65536, like XP)
     "\x01\x02\x03\x04"	// Session Key
@@ -433,6 +474,7 @@ enum {
   STATUS_INVALID_HANDLE=0xC0000008,
   ERROR_NO_MEMORY=0xc0000017,
   ERROR_ACCESS_DENIED=0xC0000022,
+  STATUS_OBJECT_NAME_INVALID=0xC0000033,
   ERROR_OBJECT_NAME_NOT_FOUND=0xc0000034,
   ERROR_NOT_SUPPORTED=0xc00000bb,
   ERROR_NETWORK_ACCESS_DENIED=0xc00000ca,
@@ -473,16 +515,34 @@ static size_t utf16toutf8(char* dest,size_t dsize,uint16_t* src,size_t ssize) {
   return dsize-Y;
 }
 
-static size_t utf8toutf16(char* dest,size_t dsize,char* src,size_t ssize) {
-  size_t X,Y;
-  char* x,* y;
-  x=src;
-  y=dest;
-  X=ssize;
-  Y=dsize;
-  memset(dest,0,dsize);
-  if (iconv(utf82wc2,&x,&X,&y,&Y)) return 0;
-  return dsize-Y;
+static size_t utf8toutf16(char* dest,size_t dsize,const char* src) {
+  /* src is a filename, so it might be either latin1 or utf8.
+     try to parse it as utf8, but if we fail on a char, we assume it's latin1 */
+  size_t c;
+  char* orig=dest;
+  while (*src) {
+    if ((src[0]&0x80)==0) {
+      c=src[0];
+      ++src;
+    } else if ((src[0]&0xe0)==0xc0 && (src[1]&0xc0)==0x80) {
+      c=((src[0]&0x1f) << 6) + (src[1]&0x3f);
+      src+=2;
+    } else if ((src[0]&0xf0)==0xe0 && (src[1]&0xc0)==0x80 && (src[2]&0xc0)==0x80) {
+      c=((src[0]&0xf) << 12) + ((src[1]&0x3f) << 6) + (src[2]&0x3f);
+      src+=3;
+    } else {
+      /* we don't support longer UTF-8 sequences because you'd need more
+       * than the 16 bits Windows has for them. */
+      c=(unsigned char)src[0];
+      ++src;
+    }
+    if (dsize<2) return 0;
+    dest[0]=c&0xff;
+    dest[1]=(c>>8);
+    dest+=2;
+    dsize-=2;
+  }
+  return dest-orig;
 }
 
 
@@ -633,13 +693,6 @@ static int smb_handle_CreateAndX(struct http_data* h,const unsigned char* c,size
     "\x00\x00";	// byte count 0
 
   if (len<2*24 || c[0]!=24) return -1;
-  /* see if it is an open for reading */
-  if ((c[16]&7)!=1) {
-    /* we only support read access */
-//    printf("non-read-access requested: %x!\n",uint32_read(c+16));
-    set_smb_error(sr,ERROR_ACCESS_DENIED,0xa2);
-    return 0;
-  }
   /* now look at file name */
   {
     size_t fnlen=uint16_read((char*)c+6);
@@ -651,6 +704,17 @@ static int smb_handle_CreateAndX(struct http_data* h,const unsigned char* c,size
     if (uint16_read((char*)c+0x31)<fnlen) return -1;
     if (fnlen%2) --fnlen;
     if (fnlen>2046 || ((uintptr_t)remotefilename%2)) return -1;
+    if (fnlen==14 && !memcmp(remotefilename,"\\\x00s\x00r\x00v\x00s\x00v\x00""c\x00",14)) {
+      set_smb_error(sr,ERROR_OBJECT_NAME_NOT_FOUND,0xa2);
+      return 0;
+    }
+    /* see if it is an open for reading */
+    if ((c[16]&7)!=1) {
+      /* we only support read access */
+  //    printf("non-read-access requested: %x!\n",uint32_read(c+16));
+      set_smb_error(sr,ERROR_ACCESS_DENIED,0xa2);
+      return 0;
+    }
     hdl=alloc_handle(&h->h);
     if (!hdl) {
 //      printf("could not open file handle!");
@@ -733,6 +797,7 @@ static int smb_handle_ReadAndX(struct http_data* h,unsigned char* c,size_t len,u
   hdl->cur=uint32_read((char*)c+7);
   if (c[0]==12)
     hdl->cur |= ((unsigned long long)uint32_read((char*)c+21))<<32;
+
 #if 0
   relofs=uint32_read((char*)c+7);
 
@@ -753,13 +818,19 @@ static int smb_handle_ReadAndX(struct http_data* h,unsigned char* c,size_t len,u
     count=mymax(uint16_read((char*)c+13),uint16_read((char*)c+11));
   if (count>65500) count=65500;
 
+#if 0
+  printf("read: %i bytes from ofs %llu (file has %llu bytes)\n",count,hdl->cur,hdl->size);
+#endif
+
   if (hdl->cur>hdl->size)
     count=0;
   else
     if (count>hdl->size-hdl->cur) count=hdl->size-hdl->cur;
 
   oldused=sr->used;
-  if (!(x=add_smb_response2(sr,nr,12*2+3,0x2e))) return -1;
+  if (!(x=add_smb_response2(sr,nr,12*2+3,0x2e))) {
+    return -1;
+  }
   uint16_pack(x+OFS16(nr,"w0"),0);	// no andx for read
   {
     off_t rem=hdl->size-hdl->cur-count;
@@ -781,12 +852,14 @@ static int smb_handle_ReadAndX(struct http_data* h,unsigned char* c,size_t len,u
 
 static int smb_handle_Trans(unsigned char* c,size_t len,struct smb_response* sr) {
   /* windows 7 calls this when trying to copy a file via cmd.exe copy */
-  /* sambs replies STATUS_NOT_SUPPORTED.  works for me. */
+  /* samba replies STATUS_NOT_SUPPORTED.  works for me. */
+#if 0
   if (len<0x34 || c[0]!=23) return -1;
   if (uint16_read(c+0x25)!=2) return -1;	/* not ioctl */
+#endif
   /* we don't really care what ioctl they were trying to call */
   /* always return the same canned answer */
-  set_smb_error(sr,0xc00000bb,0xa0);
+  set_smb_error(sr,ERROR_NOT_SUPPORTED,0xa0);
   return 0;
 }
 
@@ -832,6 +905,7 @@ static int smb_handle_Trans2(struct http_data* h,unsigned char* c,size_t len,uin
       if (hdl->filename) {
 	fnlen=hdl->filename[0];
 	filename=hdl->filename+1;
+	if (fnlen && filename[fnlen-1]==0 && filename[fnlen-2]==0) fnlen-=2;
       }
       loi=uint16_read((char*)c-smbheadersize+paramofs+2);
     } else if (subcommand==5) {
@@ -933,7 +1007,7 @@ filenotfound:
 	  uint32_pack(buf+64,0);
 	} else if (loi==0x102) {
 	  uint16_pack(buf+21,datacount+5);
-	  uint64_pack(buf+28,(unsigned long long)ss.st_blocks*ss.st_blksize);
+	  uint64_pack(buf+28,(unsigned long long)ss.st_blocks*512);
 	  uint64_pack(buf+28+8,ss.st_size);
 	  uint32_pack(buf+28+8+8,ss.st_nlink);
 	  buf[28+8+8+4]=0;
@@ -946,7 +1020,7 @@ filenotfound:
     case 0x0107:	// SMB_QUERY_FILE_ALL_INFO
       {
 	char* buf;
-	size_t datacount=78+fnlen;
+	size_t datacount=72+fnlen;
 	buf=alloca(20+100+datacount);
 	byte_copy(buf,21,
 	  "\x0a"		// word count
@@ -974,14 +1048,14 @@ filenotfound:
 	uint64_pack_ntdate(buf+28+8+8,ss.st_mtime);
 	uint64_pack_ntdate(buf+28+8+8+8,ss.st_mtime);
 	uint32_pack(buf+60,attr);	// normal file
-	uint64_pack(buf+68,(unsigned long long)ss.st_blocks*ss.st_blksize);
+	uint64_pack(buf+68,(unsigned long long)ss.st_blocks*512);
 	uint64_pack(buf+76,ss.st_size);
 	uint32_pack(buf+84,ss.st_nlink);
 	byte_zero(buf+88,8);
 	uint32_pack(buf+96,fnlen);
 	if (fnlen)
 	  byte_copy(buf+100,fnlen,filename);
-	return add_smb_response(sr,buf,21+datacount,0x32);
+	return add_smb_response(sr,buf,28+datacount,0x32);
       }
     default:
       set_smb_error(sr,ERROR_ACCESS_DENIED,0x32);
@@ -989,10 +1063,10 @@ filenotfound:
     }
   } else if (subcommand==1 ||	// FIND_FIRST2
 	     subcommand==2) {	// FIND_NEXT2
-    size_t i,l,rl;
+    size_t i,l,rl=0;
     size_t maxdatacount,sizeperrecord=0;
     char* globlatin1,* globutf8;
-    uint16_t attr,flags;
+    uint16_t attr;
     uint16_t* resume;
     DIR* d;
 #if 0
@@ -1010,7 +1084,6 @@ outofmemory:
       return -1;		// need at least six chars for "/*" in unicode
     maxdatacount=uint16_read((char*)c+7);
     attr=uint16_read((char*)c-smbheadersize+paramofs);
-    flags=uint16_read((char*)c-smbheadersize+paramofs+4);
     if (subcommand==1)
       loi=uint16_read((char*)c-smbheadersize+paramofs+6);
     else
@@ -1025,7 +1098,7 @@ outofmemory:
       sizeperrecord=0x44;
     if (subcommand==1) {
       filename=(uint16*)(c-smbheadersize+paramofs+12);
-      l=paramcount-12;
+      l=(paramcount-12)/2;
       if ((h->ftppath=malloc(l+l+4))) {
 	memcpy(h->ftppath+2,filename,l+l+2);
 	((uint16*)h->ftppath)[0]=l;
@@ -1088,12 +1161,17 @@ outofmemory:
 
     if ((uintptr_t)filename % 2)
       goto filenotfound;
-    if (filename[l-1])
+    if (l==0 || filename[l-1])
       return -1;		// want null terminated filename
     if (uint16_read((char*)&filename[l-1])=='\\' || uint16_read((char*)&filename[l-1])=='/')
       goto filenotfound;	// can't glob if filename ends in \ or /
     if (uint16_read((char*)&(filename[0]))!='\\')
       goto filenotfound;
+    for (i=0; i+2<l; ++i)
+      if (uint16_read((char*)&filename[i])<0x1f) {
+	set_smb_error(sr,STATUS_OBJECT_NAME_INVALID,0x32);
+	return 0;
+      }
     for (i=l; i>0; --i)
       if (uint16_read((char*)&filename[i])=='\\') {
 	filename[i]=0;
@@ -1104,7 +1182,18 @@ outofmemory:
       goto filenotfound;
     filename+=i+1; l-=i;
     globlatin1=alloca(l+1);
-    if (utf16tolatin1(globlatin1,l-i+1,filename,(l+1)*2)) {
+
+#if 0
+    {
+      int j;
+      printf("convert \"");
+      for (j=0; filename[j]; ++j)
+	printf("%c",filename[j]);
+      printf("\" (source size: %d, dest size %d)\n",(l+1)*2,l+1);
+    }
+#endif
+
+    if (utf16tolatin1(globlatin1,l+1,filename,(l+1)*2)) {
       globlatin1=0;
 //      puts("could not convert glob expression to latin1!");
     } // else
@@ -1138,8 +1227,8 @@ outofmemory:
 //	  printf("matching %s vs %s\n",globlatin1,de->d_name);
 //	if (globutf8)
 //	  printf("matching %s vs %s\n",globutf8,de->d_name);
-	if ((globlatin1 && !fnmatch(globlatin1,de->d_name,0)) ||
-	    (globutf8 && !fnmatch(globutf8,de->d_name,0))) {
+	if ((globlatin1 && globmatch(globlatin1,de->d_name)) ||
+	    (globutf8 && globmatch(globutf8,de->d_name))) {
 	  if (stat(de->d_name,&ss)==-1) continue;
 //	  printf("globbed ok: %s\n",de->d_name);
 	  if (!base) {
@@ -1185,8 +1274,8 @@ outofmemory:
 	    cur=base=sr->buf+sr->used;
 	    max=sr->buf+sr->allocated;
 	  }
-	  if (max-cur < 0x60 +strlen(de->d_name)*2 ||
-	      !(actualnamelen=utf8toutf16(cur+sizeperrecord,max-cur-sizeperrecord,de->d_name,strlen(de->d_name)))) {
+	  if (max-cur < 100+0x60 +strlen(de->d_name)*2 ||
+	      !(actualnamelen=utf8toutf16(cur+sizeperrecord,max-cur-sizeperrecord,de->d_name))) {
 	    // not enough space!  abort!  abort!
 	    if (subcommand==1)
 	      trans2[28]=0;
@@ -1224,7 +1313,7 @@ outofmemory:
 	    uint64_pack_ntdate(cur,ss.st_mtime); cur+=8;
 	    uint64_pack_ntdate(cur,ss.st_mtime); cur+=8;
 	    uint64_pack(cur,ss.st_size); cur+=8;
-	    uint64_pack(cur,ss.st_blksize*ss.st_blocks); cur+=8;
+	    uint64_pack(cur,512*ss.st_blocks); cur+=8;
 	    uint32_pack(cur,S_ISDIR(ss.st_mode)?0x10:0x80); cur+=4;
 	    uint32_pack(cur,actualnamelen); cur+=4;
 	    uint32_pack(cur,0); cur+=4;	// ea list length
@@ -1254,7 +1343,7 @@ outofmemory:
 	uint16_pack(trans2+9,trans2+20-sr->buf);
 	uint16_pack(trans2+13,cur-base);
 	uint16_pack(trans2+15,base-smbhdr);
-	uint16_pack(trans2+21,base-smbhdr-12);
+	uint16_pack(trans2+21,cur-base+13);
 	if (subcommand==1) {
 	  uint16_pack(trans2+26,searchcount);	// search count...!?
 	  uint16_pack(trans2+32,last-base);
@@ -1273,7 +1362,11 @@ outofmemory:
     return -1;
 }
 
-static int smb_handle_Close(struct http_data* h,unsigned char* c,size_t len,uint32_t pid,struct smb_response* sr) {
+static int smb_handle_close2(unsigned char* c,size_t len,struct smb_response* sr) {
+  return add_smb_response(sr,"\x00\x00\x00",3,0x52);
+}
+
+static int smb_handle_Close(struct http_data* h,unsigned char* c,size_t len,struct smb_response* sr) {
   struct handle* hdl;
   if (len<2*3 || c[0]!=3) return -1;
   if (!(hdl=deref_handle(&h->h,uint16_read((char*)c+1)))) {
@@ -1284,9 +1377,45 @@ static int smb_handle_Close(struct http_data* h,unsigned char* c,size_t len,uint
   return add_smb_response(sr,"\x00\x00\x00",3,0x4);
 }
 
+static void fmt_dostime(char* d,time_t t) {
+  struct tm lt;
+  localtime_r(&t,&lt);
+  /* date + time as 16-bit values */
+  uint16_pack(d,lt.tm_mday + ((lt.tm_mon+1)<<5) + ((lt.tm_year-80)<<9));
+  uint16_pack(d+2,(lt.tm_sec/2) + (lt.tm_min<<5) + (lt.tm_hour<<11));
+}
+
+static int smb_handle_Query_Information2(struct http_data* h,unsigned char* c,size_t len,struct smb_response* sr) {
+  struct handle* hdl;
+  struct stat ss;
+  char buf[100];
+  if (len<2 || c[0]!=1) return -1;
+  if (!(hdl=deref_handle(&h->h,uint16_read((char*)c+1)))) {
+    set_smb_error(sr,STATUS_INVALID_HANDLE,0x23);
+    return 0;
+  }
+  if (fstat(hdl->fd,&ss)==-1) {
+    /* can't happen */
+    set_smb_error(sr,ERROR_OBJECT_NAME_NOT_FOUND,0x23);
+    return 0;
+  }
+  buf[0]=12;	/* word count */
+  /* ctime, atime, mtime, size, allocation size, attributes (2), byte count */
+  fmt_dostime(buf+1,ss.st_ctime);
+  fmt_dostime(buf+1+4,ss.st_atime);
+  fmt_dostime(buf+1+4+4,ss.st_mtime);
+  uint32_pack(buf+1+4+4+4,ss.st_size);
+  uint32_pack(buf+1+4+4+4+4,ss.st_blocks*512);
+  buf[1+4+4+4+4+4]=1+(S_ISDIR(ss.st_mode)?0x10:0);	/* the 1 is for read-only */
+  buf[1+4+4+4+4+4+1]=0;
+  buf[1+4+4+4+4+4+2]=0;
+  buf[1+4+4+4+4+4+3]=0;
+  return add_smb_response(sr,buf,1+4+4+4+4+4+5,0x23);
+}
+
 static int smb_handle_QueryDiskInfo(unsigned char* c,size_t len,struct smb_response* sr) {
   struct statvfs sv;
-  char buf[11];
+  char buf[13];
   unsigned long long l,k;
   size_t i;
   if (len<3 || c[0]!=0) return -1;
@@ -1320,8 +1449,8 @@ static int smb_handle_QueryDiskInfo(unsigned char* c,size_t len,struct smb_respo
       uint16_pack(buf+3,1<<(i-15));
     }
   }
-  uint16_pack(buf+9,0);
-  return add_smb_response(sr,buf,11,0x80);
+  uint32_pack(buf+9,0);
+  return add_smb_response(sr,buf,13,0x80);
 }
 
 int smbresponse(struct http_data* h,int64 s) {
@@ -1340,10 +1469,19 @@ int smbresponse(struct http_data* h,int64 s) {
     /* if it's good enough for samba, it's good enough for me. */
     return 0;
   len=uint32_read_big((char*)c)&0xffffff;
-  if (len<smbheadersize) return 0;
+  if (len<smbheadersize) {
+//    printf("netbios len (%u) < smbheadersize (%u)\n",len,smbheadersize);
+    return 0;
+  }
 
-  if (validate_smb_packet(c+netbiosheadersize,len)==-1)
+//  hexdump(c,len+netbiosheadersize);
+
+  if (validate_smb_packet(c+netbiosheadersize,len)==-1) {
+//    printf("invalid smb packet!\n");
+//    hexdump(c,len+netbiosheadersize);
+//    validate_smb_packet(c+netbiosheadersize,len);
     return -1;
+  }
 
   /* is it a request?  Discard replies. */
   if (c[13]&0x80) return 0;
@@ -1361,7 +1499,7 @@ int smbresponse(struct http_data* h,int64 s) {
     switch (andxtype) {
     case 0x04:
       /* Close Request */
-      if (smb_handle_Close(h,c+cur,len-cur,uint16_read((char*)smbheader+0x1a),&sr)==-1)
+      if (smb_handle_Close(h,c+cur,len-cur,&sr)==-1)
 	goto kaputt;
       break;
 
@@ -1369,6 +1507,12 @@ int smbresponse(struct http_data* h,int64 s) {
       /* Check Directory Request */
       break;
 
+    case 0x23:
+      /* Query Information2 */
+      if (smb_handle_Query_Information2(h,c+cur,len-cur,&sr)==-1)
+	goto kaputt;
+      break;
+
     case 0x2b:
       if (smb_handle_echo(c+cur,len-cur,&sr)==-1)
 	goto kaputt;
@@ -1392,6 +1536,12 @@ int smbresponse(struct http_data* h,int64 s) {
 	goto kaputt;
       break;
 
+    case 0x52:
+      /* Find Close2 */
+      if (smb_handle_close2(c+cur,len-cur,&sr)==-1)
+	goto kaputt;
+      break;
+
     case 0x71:
       /* Tree Disconnect Request */
       if (smb_handle_TreeDisconnect(c+cur,len-cur,&sr)==-1)
@@ -1430,7 +1580,10 @@ int smbresponse(struct http_data* h,int64 s) {
     case 0xa2:
       if (smb_handle_CreateAndX(h,c+cur,len-cur,uint16_read((char*)smbheader+0x1a),&sr)==-1)
 	goto kaputt;
+      break;
 
+    default:
+      set_smb_error(&sr,ERROR_NOT_SUPPORTED,andxtype);
     }
     if (!hasandx(andxtype)) break;
     andxtype=c[cur+1];
